import ast
import os
import time
from traceback import print_exc
from termcolor import colored

from .get_model import download_model
from .filters import choose_filter_helper
from .utils import err_400_if_true, err_404_if_true, err_500_if_true, err_501_if_true, err_if_true, parse_args


args_message = {
    'batchfile': {
        'args_format': '<Comma Separated File> <Destination Path> --custom-filter=<Path to filter file (optional)> --filter=<tags (optional)> --max-images=<default=3 (optional)',
        'example': '~/lora-to-download.txt ~/organized_loras --custom-filter=example.py'
    },
    'batchstr': {
        'args_format': '<Comma Separated File> <Destination Path> --custom-filter=<Path to filter file (optional)> --filter=<tags (optional)> --max-images=<default=3 (optional)',
        'example': '"123456,78901,23456" ~/organized_loras --custom-filter=example.py'
    }
}


def get_input_strings(type: str, source: str):
    string = source
    if type == 'batchfile':
        with open(source, 'r') as file:
            string = file.read().strip()
    res = [input_str for input_str in string.replace(
        '\n', '').split(',') if input_str.strip() != '']
    return res


def batch_download(type: str, argv: list[str]):
    """Accepted type are batchfile and batchstr"""
    err_500_if_true(type != 'batchfile' and type != 'batchstr',
                    f'Unknown batch type --> {type}')

    message = args_message[type]
    kwargs, args = parse_args(argv)
    err_400_if_true(len(args) < 2,
                    f'Missing arguments. Arguments are the following: {message["args_format"]}.\nExample: {message["example"]}')
    err_404_if_true(type != 'batchstr' and not os.path.exists(
        args[0]), 'Source directory does not exist.')

    input_li = None
    input_li = get_input_strings(type, args[0])

    err_500_if_true(input_li == None, 'input_li is of type None')
    err_500_if_true(len(input_li) == None,
                    f'get_input_strings was not able to parse argument, {args[0]}')

    filter_model = choose_filter_helper(kwargs)
    err_500_if_true(filter_model == None,
                    'Error: filter_model is of type None')

    for input_str in input_li:
        try:
            download_model(
                input_str=input_str,
                create_dir_path=filter_model,
                dst_root_path=args[1],
                download_image=True,
                max_img_count=(kwargs['max-images'] if 'max-images' in kwargs else 3))
            time.sleep(2)
        except Exception as e:
            try:
                res = ast.literal_eval(str(e))
                print(
                    f"{colored('Status ' + str(res['status']), 'red', attrs=['bold'])}: {colored(res['message'], 'red')}")
            except:
                # TODO: What do we do about error status from civitai's server?
                print_exc(e)



#!/usr/bin/env python3

import ast
import sys
from traceback import print_exc
from termcolor import colored

from .utils import err_400_if_true
from .batch_download import batch_download


def main():
    try:
        available_actions = ['batchfile', 'batchstr']
        argv = sys.argv[1:]
        err_400_if_true(len(argv) == 0,
                        f'Error: Need to provide arguments -> <action> *<action\'s arguments>.\nAvailable actions: {available_actions}')

        if argv[0] == 'batchfile':
            batch_download('batchfile', argv[1:])
        elif argv[0] == 'batchstr':
            batch_download('batchstr', argv[1:])
        else:
            err_400_if_true(
                True, TypeError, f'Error: Unknown action.\nAvailable actions: {available_actions}')
    except Exception as e:
        try:
            res = ast.literal_eval(str(e))
            print(
                f"{colored('Status ' + str(res['status']), 'red', attrs=['bold'])}: {colored(res['message'], 'red')}")
        except:
            # TODO: What do we do about error status from civitai's server?
            print_exc(e)



import json
from typing import Callable
from tqdm import tqdm


def write_to_file(path, content, mode: str = None):
    f = open(path, mode) if mode != None else open(path, 'w')
    f.write(content)
    f.close()


def write_to_file_with_progress_bar(path, res, mode: str = None):
    """Stream must have been enabled -> request.get(url, stream=True)"""
    total_size_in_bytes = int(res.headers.get('content-length', 0))
    block_size = 1024
    progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
    with open(path, mode if mode != None else 'w') as file:
        for data in res.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    err_500_if_true(total_size_in_bytes != 0 and progress_bar.n !=
                    total_size_in_bytes, 'Unexpected error while writing file with progress bar.')


def find_in_list(li, cond_fn: Callable[[any, int], bool], default=None):
    """Given a list and a condition function, where the first argument is the index of the item and the second argument is the value of the item, it returns the first value in the list when the condition is true"""
    return next((item for i, item in enumerate(li) if cond_fn(item, int)), default)


def parse_args(args: list[str]):
    """Returns a tuple of (kwargs, args)"""
    kwargs = {}
    non_kwargs = []

    for arg in args:
        if arg.startswith('--'):
            key_value = arg[2:].split('=')
            if len(key_value) == 2:
                key, value = key_value
                kwargs[key] = value
        else:
            non_kwargs.append(arg)

    return (kwargs, non_kwargs)


def err_if_true(condition: bool, message: str, status: int):
    """Raises an exception fi condition is true"""

    if condition:
        dic = {
            "status": status,
            "message": message
        }
        raise Exception(json.dumps(dic))
    else:
        return False


def err_400_if_true(condition: bool, message: str):
    return err_if_true(condition, message, 400)


def err_404_if_true(condition: bool, message: str):
    return err_if_true(condition, message, 404)


def err_500_if_true(condition: bool, message: str):
    return err_if_true(condition, 'InternalError: ' + message, 500)


def err_501_if_true(condition: bool, message: str):
    return err_if_true(condition, 'ImplementationError: ' + message, 501)



from json import dumps
import requests
import os
import re
from typing import Callable, Dict, List

from termcolor import colored

from .utils import err_400_if_true, err_404_if_true, err_500_if_true, err_501_if_true, err_if_true, write_to_file, write_to_file_with_progress_bar


class Metadata:
    __id_string = None

    model_name = None
    model_id = None
    version_id = None

    model_dict = None
    version_dict = None

    download_url = None
    images_dict_li = None
    nsfw = False

    def __init__(self, id_string: str):
        self.__id_string = id_string

        if id_string.isdigit():
            self.__id_handler()
        elif "/api/" in id_string:
            self.__api_url_handler()
        else:
            self.__url_handler()

    def __id_handler(self):
        id = [self.__id_string]
        self.__handler(id, 'id')

    def __url_handler(self):
        model_id_regex = r'(?<=models\/)\d+'
        version_id_regex = r'(?<=modelVersionId=)\d+'
        model_id = re.search(model_id_regex, self.__id_string)
        version_id = re.search(version_id_regex, self.__id_string)
        err_400_if_true(model_id == None,
                        f'Incorrect format for the url/id provided: {self.__id_string}')
        model_id = model_id.group(0)
        err_500_if_true(
            model_id == None, f'Unknown error while parsing model id for the url/id provided: {self.__id_string}')

        if version_id:
            version_id = version_id.group(0)
            err_500_if_true(
                version_id == None, f'Unknown error while parsing version id for the url/id provided: {self.__id_string}')
            self.__handler([model_id, version_id], 'url')
        else:
            self.__handler([model_id], 'url')

    def __api_url_handler(self):
        version_id_regex = r'(?<=models\/)\d+'
        version_id = re.search(version_id_regex, self.__id_string)
        err_400_if_true(version_id == None,
                        f'Incorrect format for the url provided: {self.__id_string}')
        version_id = version_id.group(0)
        err_500_if_true(
            version_id == None, f'Unknown error while parsing version id for the url provided: {self.__id_string}')

        self.__handler([version_id], 'api')

    def __handler(self, ids: List[int], type: str):
        if type == 'url' and len(ids) == 2:
            self.model_id = ids[0]
            self.version_id = ids[1]
            self.model_dict = self.__get_model_metadata()
            self.version_dict = self.__get_version_metadata()
        elif type == 'id' or (type == 'url' and len(ids) == 1):
            self.model_id = ids[0]
            self.model_dict = self.__get_model_metadata()
            err_404_if_true(len(self.model_dict['modelVersions']) == 0,
                            f'No model versions found from model id, {self.model_id}')
            self.version_dict = self.model_dict['modelVersions'][0]
            self.version_id = str(self.version_dict['id'])
        elif type == 'api':
            self.version_id = ids[0]
            self.version_dict = self.__get_version_metadata()
            self.model_id = str(self.version_dict['modelId'])
            self.model_dict = self.__get_model_metadata()
        else:
            err_400_if_true(True,
                            f'Incorrect format sent ({ids}, {type}): "{self.__id_string}"')

        self.download_url = self.version_dict['downloadUrl']
        self.images_dict_li = self.version_dict['images']
        self.nsfw = self.model_dict['nsfw']
        self.model_name = self.model_dict['name']

    def __get_model_metadata(self):
        """Returns json object if request succeeds, else print error and returns None"""
        metadata_url = 'https://civitai.com/api/v1/models/' + self.model_id
        return self.__get_metadata(metadata_url)

    def __get_version_metadata(self):
        metadata_url = 'https://civitai.com/api/v1/model-versions/' + self.version_id
        return self.__get_metadata(metadata_url)

    def __get_metadata(self, url: str):
        meta_res = requests.get(url)
        err_if_true(meta_res.status_code != 200,
                    f'Downloading metadata from CivitAI for "{self.__id_string}" failed when trying to request metadata from "{url}"', meta_res.status_code)
        return meta_res.json()


def _download_image(dirpath: str, images: list[Dict], nsfw: bool, max_img_count):
    image_urls = []

    for dict in images:
        if len(image_urls) == max_img_count:
            break
        if not nsfw and dict['nsfw'] != 'None':
            continue
        else:
            image_urls.append(dict['url'])

    for url in image_urls:
        image_res = requests.get(url)
        err_if_true(image_res.status_code != 200,
                    f'Downloading image from CivitAI failed for the url: {url}', image_res.status_code)
        write_to_file(os.path.join(
            dirpath, os.path.basename(url)), image_res.content, 'wb')


def _get_filename_and_model_res(input_str: str, metadata: Metadata):
    # Request model
    res = requests.get(metadata.download_url, stream=True, headers={
        "Accept-Charset": "utf-16"})
    err_if_true(res.status_code != 200,
                f'Downloading model from CivitAI failed for model id, {metadata.model_id}, and version id, {metadata.version_id}', res.status_code)

    # Find filename from content disposition
    content_disposition = res.headers.get('Content-Disposition')
    err_404_if_true(content_disposition == None,
                    f'Downloaded model from CivitAI has no content disposition header available.')
    filename = None

    try:
        filename = content_disposition.split(
            'filename=')[-1].strip('"').encode('latin-1').decode('utf-8')
    except UnicodeDecodeError:
        # Alternative solution for finding filename
        for file in metadata.version_dict['files']:
            file_version_id = re.search(
                r'(?<=models\/)\d+', file['downloadUrl'])
            if file_version_id != None and file_version_id.group(0) == metadata.version_id:
                filename = file['name']
                break

    if filename == None:
        err_500_if_true(filename == None,
                        f"Error: Unable to retrieve filename for {input_str}")

    return (res, filename)


def download_model(input_str: str, create_dir_path: Callable[[Dict, Dict, str, str], str], dst_root_path: str, download_image: bool, max_img_count: int):
    """
        Downloads the model's safetensors and json metadata files.
        create_dir_path is a callback function that takes in the following: metadata dict, specific model's data as dict, filename, and root path.
    """
    err_500_if_true(input_str == None or create_dir_path == None or dst_root_path ==
                    None or download_image == None or max_img_count == None, 'download_model received a None type in one of its parameter')

    metadata = Metadata(input_str)

    print(colored(
        f"""Now downloading \"{metadata.model_name}\"...
            - Model ID: {metadata.model_id}
            - Version ID: {metadata.version_id}\n""",
        'blue'))

    model_res, filename = _get_filename_and_model_res(input_str, metadata)

    # Create empty directory recursively
    filename_without_ext, filename_ext = os.path.splitext(filename)
    dst_dir_path = create_dir_path(
        metadata.model_dict, metadata.version_dict, filename_without_ext, dst_root_path)
    if not os.path.exists(dst_dir_path):
        os.makedirs(dst_dir_path)

    # Write model to the directory
    json_path = os.path.join(
        dst_dir_path, f'{filename_without_ext}-mid_{metadata.model_id}-vid_{metadata.version_id}.json')
    model_path = os.path.join(
        dst_dir_path, f'{filename_without_ext}-mid_{metadata.model_id}-vid_{metadata.version_id}{filename_ext}')
    write_to_file(json_path, dumps(
        metadata.model_dict, indent=2, ensure_ascii=False))
    write_to_file_with_progress_bar(model_path, model_res, 'wb')
    if download_image:
        _download_image(
            dst_dir_path, metadata.version_dict['images'], metadata.nsfw, max_img_count)

    print(colored(
        f"""\nDownload completed for \"{metadata.model_name}\" 
            - Model ID: {metadata.model_id}
            - Version ID: {metadata.version_id}
            - Path: {model_path}\n""", 'green'))

    print('---------------------------\n')



import importlib.util
from typing import Dict
import os


def create_basic_path(model_dict: Dict, version_dict: Dict, filename: str, root_path: str):
    """This is useful when you only just want all of the lora in the same parent folder."""
    return os.path.join(root_path, filename)


def create_path_by_tags(model_dict: Dict, version_dict: Dict, filename: str, root_path: str):
    """Create nested directories with info about the specific model given root path and the model dict. If no tag is found in the model's json, the default dir path name used will be the last element of twodim_tags' inner list."""
    twodim_tags = [
        ['anime', 'non-anime'],
        ['style', 'poses', 'clothings', 'character',
         'celebrity', 'concept', 'unknown']
    ]

    model_tags = model_dict['tags']
    path = os.path.join(root_path, version_dict['baseModel'].replace(' ', '_'))

    for tags in twodim_tags:
        matched_tags = list(set(tags) & set(model_tags))
        path = os.path.join(
            path, tags[-1] if len(matched_tags) == 0 else matched_tags[0])

    path = os.path.join(path, filename)

    return path


def choose_filter_helper(kwargs: Dict[str, str]):
    filter_model = None
    if 'custom-filter' in kwargs:
        if not os.path.exists(kwargs['custom-filter']):
            return print('Error: Custom filter python file does not exist')
        spec = importlib.util.spec_from_file_location(
            'plugin', kwargs['custom-filter'])
        plugin = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(plugin)
        filter_model = plugin.filter_model
    elif 'filter' in kwargs:
        if kwargs['filter'] == 'tags':
            filter_model = create_path_by_tags
        elif kwargs['filter'] == 'basic':
            filter_model = create_basic_path
        else:
            return print(f'Error: Unknown filter specified. The available built-in filters are {["tags", "basic"]} (basic is the default filter)')
    else:
        filter_model = create_basic_path
    return filter_model
